# Copyright 3D Control Systems, Inc. All Rights Reserved 2017-2025.
# Built in San Francisco.

# This software is distributed under a commercial license for personal,
# educational, corporate or any other use.
# The software as a whole or any parts of it is prohibited for distribution or
# use without obtaining a license from 3D Control Systems, Inc.

# All software licenses are subject to the 3DPrinterOS terms of use
# (available at https://www.3dprinteros.com/terms-and-conditions/),
# and privacy policy (available at https://www.3dprinteros.com/privacy-policy/)

import ipaddress
import json
import logging
import socket
import pprint


import requests
import findssh
import findssh.threadpool
import log


class HTTPDetector:

    PORT = 80
    TIMEOUT = 1
    FINDSSH_SERVICE = "http"
    PATH = "/"
    RESPONSE_FIELD_NAME = ""
    RESPONSE_TEXT_MATCH = ""
    METHOD = "GET"
    REQ_BODY = None
    REQ_HEADERS = None
    RESP_MATCH_IS_SNR = False
    PASS_CACHE_ATTR_NAME = 'cached_http_hosts'

    def __init__(self, parent=None, profile=None):
        self.logger = logging.getLogger(__name__)
        self.ip = None
        if parent:
            self.ip = ipaddress.ip_address(parent.local_ip)
        if not self.ip:
            self.ip = findssh.getLANip()
        self.parent = parent
        self.profile = profile
        self.port = self.PORT
        self.cached_http_hosts = []
        self.discovered_printers = []

    def process_hosts(self, hosts):
        matched_hosts = {}
        for host in hosts:
            host_ip = str(host[0])
            if self.parent_stop():
                break
            match = self.check_ip(host_ip)
            if match:
                matched_hosts[host_ip] = match
        return matched_hosts

    def request(self, host):
        if self.port in (443, 8443):
            proto = "https://"
        else:
            proto = "http://"
        url = proto + host
        if self.port not in (80, 443):
            url += ":" + str(self.port)
        if self.PATH:
            url += self.PATH
        kwargs = {}
        if self.REQ_BODY:
            kwargs['data'] = self.REQ_BODY
        if self.REQ_HEADERS:
            kwargs['headers'] = self.REQ_HEADERS
        try:
            resp = requests.request(self.METHOD, url, timeout=self.TIMEOUT, verify=False, **kwargs)
            return resp.text
        except:
            pass
        finally:
            try:
                resp.close()
            except:
                pass

    def check_ip(self, host):
        resp_body = self.request(host)
        if resp_body:
            #self.logger.debug("Response: " + resp_body)
            if self.RESPONSE_FIELD_NAME:
                try:
                    decoded = json.loads(resp_body)
                    if type(decoded) != dict:
                        raise TypeError("Response is not json dictionary")
                except (ValueError, TypeError) as e:
                    pass
                    #self.logger.debug(f"HTTP detector got non json response from {host}: {e}")
                else:
                    #self.logger.debug('Response: %s', pprint.pformat(decoded))
                    return decoded.get(self.RESPONSE_FIELD_NAME)
            elif self.RESPONSE_TEXT_MATCH:
                #self.logger.debug("Searching for text match in the response from " + str(host))
                if self.RESPONSE_TEXT_MATCH in resp_body:
                    #self.logger.debug("Found")
                    return self.RESPONSE_TEXT_MATCH
                #self.logger.debug("Not found")
            else:
                self.logger.error('Invalid use of HTTPDetector. Ether RESPONSE_FIELD_NAME or RESPONSE_FIELD_NAME class attributes should be set, but none is.')

    def parent_stop(self):
        return self.parent and self.parent.stop_flag

    def detect(self, already_know_ip=None, non_default_port=None):
        if non_default_port:
            self.port = non_default_port
        else:
            self.port = self.PORT
        hosts = []
        if already_know_ip:
            hosts = [(already_know_ip, self.port)]
        elif self.cached_http_hosts:
            host = self.cached_http_hosts
        else:
            if self.ip:
                try:
                    network = findssh.netfromaddress(self.ip)
                except:
                    self.logger.error('Error on getting network from ip')
                else:
                    try:
                        hosts = findssh.runner(network, self.port, self.FINDSSH_SERVICE, self.TIMEOUT)
                    except:
                        try:
                            hosts = findssh.threadpool.get_hosts(network, self.port, self.FINDSSH_SERVICE, self.TIMEOUT)
                        except:
                            pass
                if hosts:
                    hosts = list(hosts)
                self.cached_http_hosts = hosts
        #self.logger.debug("All detected HTTP hosts: " + str(hosts))
        if self.profile:
            VID, PID = self.profile.get('vids_pids', [("XXXX","XXXX")])[0]
        else:
            VID, PID = "XXXX", "XXXX" #dummy values only for debug mode
        discovered_printers = []
        # TODO refactor all the mess with overload of response of proccess_hosts
        for ip, host_match_data in self.process_hosts(hosts).items():
            discovered_printers.append(self.parse_matched_hosts(ip, VID, PID, host_match_data))
        self.discovered_printers = discovered_printers

    def parse_matched_hosts(self, IP, VID, PID, match_data):
        #self.logger.debug("Detected: " + str(match_data))
        printer = {'VID': VID, 'PID': PID}
        if isinstance(match_data, dict):
            printer.update(match_data)
        elif isinstance(match_data, str) and self.RESP_MATCH_IS_SNR:
            printer['SNR'] = match_data
        if 'SNR' not in printer:
            printer['SNR'] = IP
        printer['IP'] = IP
        return printer

if __name__ == "__main__":

    detector = HTTPDetector(None)
    detector.detect()
    pprint.pprint(detector.discovered_printers)
