# Copyright 3D Control Systems, Inc. All Rights Reserved 2017-2019.
# Built in San Francisco.

# This software is distributed under a commercial license for personal,
# educational, corporate or any other use.
# The software as a whole or any parts of it is prohibited for distribution or
# use without obtaining a license from 3D Control Systems, Inc.

# All software licenses are subject to the 3DPrinterOS terms of use
# (available at https://www.3dprinteros.com/terms-and-conditions/),
# and privacy policy (available at https://www.3dprinteros.com/privacy-policy/)

import time
import json
import select
import socket
import logging
import threading
import operator
import os.path
import pprint

import config
import usb_detect
import paths
import http_rrfw_detect
import http_reperier_detect
import zeroconf_detect
import zeroconf_formlabs_detect
import upnp_detect
import broadcast_printer_detector
import base_detector
import http_prusalink_detect
import printer_settings_and_id


# TODO refactor
# 1) use multithreading or asyncio to scan simultaneously
# 2) decouple starting, running and processing from network detector and other detectors
# 3) refactor to simplify network detector logic and minimize this class

SCANNERS = {'broadcast': broadcast_printer_detector.BroadcastPrinterDetector,\
            'birdwing': broadcast_printer_detector.BroadcastPrinterDetector,\
            'rrfw' : http_rrfw_detect.HTTPRRFWDetector,\
            'repetier' : http_reperier_detect.HTTPRepetierDetector,\
            'zeroconf' : zeroconf_detect.ZeroconfDetector,\
            'zeroconf_formlabs' : zeroconf_formlabs_detect.ZeroconfDetector, \
            'upnp' : upnp_detect.UPNPDetector,
            'prusa' : http_prusalink_detect.HTTPPrusaLinkDetector,
            'noscanner': None,
            None: None}

class NetworkDetector(base_detector.BaseDetector):

    STORAGE_FILE_PATH = os.path.join(paths.CURRENT_SETTINGS_FOLDER, 'detected_printers.json')
    DEFAULT_SCANNER_CLASS = broadcast_printer_detector.BroadcastPrinterDetector

    PRINTER_ID_DICT_KEYS = ('VID', 'PID', 'SNR', 'IP')
    UNKNOWN_PROFILE_VID_PID = 'XXXX', 'XXXX'

    @staticmethod
    def is_same_printer(checking_dict, checked_dict): #TODO rename args in all other classes
        for key in NetworkDetector.PRINTER_ID_DICT_KEYS:
            checking_value = checking_dict.get(key)
            checked_value = checked_dict.get(key)
            if checking_value and checking_value != checked_value:
                return False
        return True

    @staticmethod
    def get_all_network_conns(profile):
        return filter(lambda conn: conn.get('type') == 'LAN', profile.get('v2', {}).get('connections', []))

    def __init__(self, parent=None):
        super().__init__(self)
        self.parent = parent
        self.lists_lock = threading.RLock()
        self.run_lock = threading.Lock()
        self.detected_printers = []
        self.printers_add_by_ip = []
        self.load_printers_list()
        caches = {}
        self.network_detector_run_flag = False

    def refresh_on_next_loop(self):
        self.network_detector_run_flag = True

    def parent_stop(self):
        return self.parent and self.parent.stop_flag

    def get_all_network_profiles(self):
        profiles = sorted([p for p in config.get_profiles() if p.get('network_detect') or any(self.get_all_network_conns(p))], key=operator.itemgetter('name'))
        return profiles

    def get_unique_network_profiles(self):
        profiles = []
        network_detects = []
        for profile in self.get_all_network_profiles():
            detector = profile.get('network_detect')
            if detector and detector not in network_detects:
                profiles.append(profile)
                network_detects.append(detector)
        return profiles

    def detect(self, overwrite_existing=True):
        self.logger.info('Scanning for network printers...')
        detected_printers = []
        caches = {}
        for profile in self.get_unique_network_profiles():
            if self.parent_stop():
                return []
            scanner_name = profile.get('network_detect', {}).get('scanner')
            if scanner_name and scanner_name != 'noscanner':
                scanner_class = SCANNERS.get(scanner_name)
                if not scanner_class:
                    self.logger.error(f'Unknown scanner name: {scanner_name}')
                    continue
                scanner = scanner_class(self.parent, profile)
                cache_attr_name = getattr(scanner_class, 'PASS_CACHE_ATTR_NAME', None)
                cache = caches.get(cache_attr_name)
                if cache_attr_name and cache:
                    setattr(scanner, cache_attr_name, cache)
                scanner.detect()
                if not cache and cache_attr_name:
                    cache = getattr(scanner, cache_attr_name, None)
                    if cache:
                        caches[cache_attr_name] = cache
                self.logger.debug('Scanner:\t%s\t%s' % (scanner_name, scanner.discovered_printers))
                if scanner.discovered_printers:
                    for printer in scanner.discovered_printers:
                        for printers_list in (self.detected_printers, self.printers_add_by_ip, detected_printers):
                            for existing_printer in printers_list:
                                if self.is_same_printer(printer, existing_printer):
                                    self.logger.info(f'Removing existing printer {existing_printer} to a printer with same is {printer}')
                                    if overwrite_existing:
                                        printers_list[printers_list.index(existing_printer)] = printer
                                        break
                            else:
                                continue
                            break
                        else:
                            detected_printers.append(printer)
        self.logger.info('Discovered printers:\n%s' % detected_printers)
        return detected_printers

    def get_printers_list(self):
        with self.run_lock:
            if self.network_detector_run_flag:
                detected = self.detect()
                with self.lists_lock:
                    for printer in detected:
                        if printer not in self.detected_printers and printer not in self.printers_add_by_ip:
                            self.detected_printers.append(printer)
                    self.save_printers_list()
                self.network_detector_run_flag = False
        return self.detected_printers + self.printers_add_by_ip

    def printers_list_with_profiles(self):
        printers_list = []
        for info in self.get_printers_list():
            for pi in getattr(self.parent, "printer_interfaces", []):
                if pi.usb_info == info:
                    printers_list.append({"usb_info": info, "profile": pi.printer_profile})
                    break
            else:
                printers_list.append({"usb_info": info, "profile": {}})
        return printers_list

    def remember_printer(self, printer_type_name_or_alias, ip, port=None, vid=None, pid=None, serial_number=None, password=None, ssh_password=None, run_detector=False, conn_id=None):
        for profile in self.get_all_network_profiles():
            if printer_type_name_or_alias == profile.get("name") or printer_type_name_or_alias == profile.get("alias"):
                printer_type_alias = profile.get("alias")
                scanner_name = profile.get('network_detect', {}).get('scanner')
                if run_detector and scanner_name: #enable scan for makerbots 5+ gen
                    if scanner_name in SCANNERS:
                        scanner_class = SCANNERS[scanner_name]
                        if scanner_class and scanner_class.CAN_DETECT_SNR:
                            scanner = scanner_class(self.parent, profile)
                            scanner.detect(already_know_ip=ip, non_default_port=port)
                            for printer in scanner.discovered_printers:
                                if vid and vid != printer['VID']:
                                    continue
                                if pid and pid != printer['PID']:
                                    continue
                                if serial_number and serial_number != printer['SNR']:
                                    continue
                                if ip is not None:
                                    printer['IP'] = ip
                                if port is not None:
                                    printer['PORT'] = port 
                                if password is not None:
                                    printer['PASS'] = password
                                self.add_printer_to_list(printer, self.printers_add_by_ip)
                                return True
                if not serial_number:
                    serial_number = ip
                if not vid or not pid:
                    if "v2" in profile and not profile.get('network_detect'):
                        conn_dicts = list(self.get_all_network_conns(profile))
                        if conn_dicts:
                            if conn_id:
                                for conn in conn_dicts:
                                    if conn.get('id') == conn_id:
                                        vid = conn['ids'][0]['VID']
                                        pid = conn['ids'][0]['PID']
                                        break
                            else:
                                try:
                                    vid = conn_dicts[0]['ids'][0]['VID']
                                    pid = conn_dicts[0]['ids'][0]['PID']
                                    conn_id = conn_dicts[0]['id']
                                except (IndexError, AttributeError, KeyError):
                                    pass
                            if not vid or not pid:
                                self.logger.warning(f'Unable to get VID and PID for {printer_type_name_or_alias}. Using v1 profile ids')
                if not vid or not pid:
                    vid, pid = profile["vids_pids"][0]
                printer = {
                    'IP': ip,
                    'SNR': serial_number,
                    'VID': vid,
                    'PID': pid
                }
                printer_settings = {'type_alias': printer_type_alias}
                if conn_id:
                    printer_settings['connection_id'] = conn_id
                printer_settings_and_id.save_settings(printer_settings_and_id.create_id_string(printer), printer_settings)
                if port is not None:
                    printer['PORT'] = port
                if password is not None:
                    printer['PASS'] = password
                if ssh_password is not None:
                    printer['SSHP'] = ssh_password
                with self.lists_lock:
                    for printers_list in (self.detected_printers, self.printers_add_by_ip):
                        for existing_printer in printers_list:
                            if self.is_same_printer(printer, existing_printer):
                                self.logger.info(f'Removing existing printer {existing_printer} to a printer with same is {printer}')
                                printers_list.remove(existing_printer)
                        self.printers_add_by_ip.append(printer)
                        self.save_printers_list()
                self.logger.info('Network printers list updated to: ' + pprint.pformat(self.printers_add_by_ip))
                return True
        return False

    def forget_printer(self, usb_info):
        with self.lists_lock:
            for group in (self.detected_printers, self.printers_add_by_ip):
                for printer in group:
                    # vid = usb_info.get('VID')
                    # pid = usb_info.get('PID')
                    # ip = usb_info.get('IP')
                    # snr = usb_info.get('SNR')
                    if self.is_same_printer(usb_info, printer):
                        group.remove(printer)
                        self.logger.info(f'Forgetting a printer with id: {usb_info}')
                        self.save_printers_list()
                        return True
            self.logger.warning(f'Unable to forget a printer with id: {usb_info}')
            return False


    def add_printer_to_list(self, printer, list_to_add, allow_overwrite=True, extra_excusion_lists=[]):
        with self.lists_lock:
            for printers_list in [self.detected_printers, self.printers_add_by_ip].extend(extra_excusion_lists):
                for existing_index, existing_printer in enumerate(printers_list):
                    indexes_to_remove = []
                    if self.is_same_printer(printer, existing_printer):
                        self.logger.info(f'Removing existing printer {existing_printer} to a printer with same is {printer}')
                        if allow_overwrite:
                            if list_to_add == printers_list:
                                printers_list[existing_index] = printer
                            else:
                                indexes_to_remove.append(index)
                for index in indexes_to_remove:
                    del printers_list[index]
            if printer not in list_to_add:               
                list_to_add.append(printer)
        self.save_printers_list()

    def save_printers_list(self):
        file_path = self.STORAGE_FILE_PATH
        try:
            with open(file_path, 'w') as f:
                data = {'detected': self.detected_printers, 'added_by_ip': self.printers_add_by_ip}
                json_config = json.dumps(data, indent = 4, separators = (',', ': '))
                f.write(json_config)
        except Exception as e:
            self.logger.error("Error writing data to %s: %s" % (file_path, str(e)))

    def load_printers_list(self):
        with self.lists_lock:
            file_path = self.STORAGE_FILE_PATH
            try:
                with open(file_path, 'r') as f:
                    data = f.read()
                    self.logger.info("Stored printers list: %s" % data)
                    printers_list = json.loads(data)
                    self.detected_printers = printers_list.get('detected', [])
                    self.printers_add_by_ip = printers_list.get('added_by_ip', [])
            except FileNotFoundError:
                pass
            except Exception as e:
                self.logger.error("Error reading data from %s: %s" % (file_path, str(e)))


if __name__ == '__main__':
    from pprint import pprint

    logging.basicConfig(level=logging.DEBUG)
    nd = NetworkDetector()
    nd.refresh_on_next_loop()
    printers = nd.get_printers_list()
    print("="*80)
    print("Network printers: ")
    pprint(printers)
